STEP_DECORATORS_DESC = [
    ("deepspeed", ".deepspeed_libs.deepspeed_decorator.DeepspeedDecorator"),
    ("huggingface_card", ".deepspeed_libs.hugging_face.card_decorator.HuggingFaceCardDecorator"),
]
__mf_promote_submodules__ = ["az_store", "hf_callbacks", "deepspeed_libs",]